@@ -39,7 +39,7 @@ class AsmAndAttributesMixin(object):
                 " : ".join(
                     self.visit(c) for c in components))
 
-    def _generate_type(self, n, modifiers=[]):
+    def _generate_type(self, n, modifiers=[], emit_declname=True):
         """ Recursive generation from a type node. n is the type node.
             modifiers collects the PtrDecl, ArrayDecl and FuncDecl modifiers
             encountered on the way down to a TypeDecl, to allow proper
@@ -54,22 +54,29 @@ class AsmAndAttributesMixin(object):
                 s += ' '.join(n.quals) + ' '
             s += self.visit(n.type)
 
-            nstr = n.declname if n.declname else ''
+            nstr = n.declname if n.declname and emit_declname else ''
             # Resolve modifiers.
             # Wrap in parens to distinguish pointer to array and pointer to
             # function syntax.
             #
             for i, modifier in enumerate(modifiers):
                 if isinstance(modifier, c_ast.ArrayDecl):
-                    if (i != 0 and isinstance(modifiers[i - 1], c_ast.PtrDecl)):
+                    if i != 0 and isinstance(modifiers[i - 1], c_ast.PtrDecl):
                         nstr = '(' + nstr + ')'
-                    nstr += '[' + self.visit(modifier.dim) + ']'
+
+                    # BUG FIX: pycparser ignores quals
+                    dim_quals = (' '.join(modifier.dim_quals) + ' '
+                                 if modifier.dim_quals else '')
+
+                    nstr += '[' + dim_quals + self.visit(modifier.dim) + ']'
+
                 elif isinstance(modifier, c_ast.FuncDecl):
-                    if (i != 0 and isinstance(modifiers[i - 1], c_ast.PtrDecl)):
+                    if i != 0 and isinstance(modifiers[i - 1], c_ast.PtrDecl):
                         nstr = '(' + nstr + ')'
                     nstr += '(' + self.visit(modifier.args) + ')'
+
                 elif isinstance(modifier, FuncDeclExt):
-                    if (i != 0 and isinstance(modifiers[i - 1], c_ast.PtrDecl)):
+                    if i != 0 and isinstance(modifiers[i - 1], c_ast.PtrDecl):
                         nstr = '(' + nstr + ')'
                     nstr += '(' + self.visit(modifier.args) + ')'
 
@@ -89,20 +96,29 @@ class AsmAndAttributesMixin(object):
                         quals = quals + ' '
                     nstr = '*' + quals + nstr
 
+            if hasattr(n, "asm") and n.asm:
+                nstr += self.visit(n.asm)
+
             if hasattr(n, "attributes") and n.attributes.exprs:
                 nstr += ' __attribute__((' + self.visit(n.attributes) + '))'
 
             if nstr:
                 s += ' ' + nstr
             return s
+
         elif typ == c_ast.Decl:
             return self._generate_decl(n.type)
+
         elif typ == c_ast.Typename:
-            return self._generate_type(n.type)
+            return self._generate_type(n.type, emit_declname=emit_declname)
+
         elif typ == c_ast.IdentifierType:
             return ' '.join(n.names) + ' '
+
         elif typ in (c_ast.ArrayDecl, c_ast.PtrDecl, c_ast.FuncDecl, FuncDeclExt):
-            return self._generate_type(n.type, modifiers + [n])
+            return self._generate_type(
+                    n.type, modifiers + [n], emit_declname=emit_declname)
+
         else:
             return self.visit(n)
 
@@ -130,14 +146,17 @@ class AsmAndAttributesMixin(object):
 
 class GnuCGenerator(AsmAndAttributesMixin, CGeneratorBase):
     def visit_TypeOfDeclaration(self, n):
-        return "__typeof__(%s)" % self.visit(n.declaration)
+        return "%s(%s)" % (n.typeof_keyword, self.visit(n.declaration))
 
     def visit_TypeOfExpression(self, n):
-        return "__typeof__(%s)" % self.visit(n.expr)
+        return "%s(%s)" % (n.typeof_keyword, self.visit(n.expr))
 
     def visit_TypeList(self, n):
         return ', '.join(self.visit(ch) for ch in n.types)
 
+    def visit_RangeExpression(self, n):
+        return '%s ... %s' % (self.visit(n.first), self.visit(n.last))
+
 
 class GNUCGenerator(GnuCGenerator):
     def __init__(self):
