@@ -16,8 +16,7 @@ class CParserBase(pycparser.c_parser.CParser):
         kwds['yacctab'] = 'pycparserext.yacctab'
         pycparser.c_parser.CParser.__init__(self, **kwds)
 
-    def parse(self, text, filename='', debuglevel=0,
-            initial_type_symbols=set()):
+    def parse(self, text, filename='', debuglevel=0, initial_type_symbols=set()):
         self.clex.filename = filename
         self.clex.reset_lineno()
 
@@ -123,7 +122,8 @@ class PreprocessorLine(c_ast.Node):
 
 
 class TypeOfDeclaration(c_ast.Node):
-    def __init__(self, declaration, coord=None):
+    def __init__(self, typeof_keyword, declaration, coord=None):
+        self.typeof_keyword = typeof_keyword
         self.declaration = declaration
         self.coord = coord
 
@@ -137,11 +137,12 @@ class TypeOfDeclaration(c_ast.Node):
         if self.declaration is not None:
             yield self.declaration
 
-    attr_names = ()
+    attr_names = ('typeof_keyword',)
 
 
 class TypeOfExpression(c_ast.Node):
-    def __init__(self, expr, coord=None):
+    def __init__(self, typeof_keyword, expr, coord=None):
+        self.typeof_keyword = typeof_keyword
         self.expr = expr
         self.coord = coord
 
@@ -155,6 +156,29 @@ class TypeOfExpression(c_ast.Node):
         if self.expr is not None:
             yield self.expr
 
+    attr_names = ('typeof_keyword',)
+
+
+class RangeExpression(c_ast.Node):
+    def __init__(self, first, last, coord=None):
+        self.first = first
+        self.last = last
+        self.coord = coord
+
+    def children(self):
+        nodelist = []
+        if self.first is not None:
+            nodelist.append(("first", self.first))
+        if self.last is not None:
+            nodelist.append(("last", self.last))
+        return tuple(nodelist)
+
+    def __iter__(self):
+        if self.first is not None:
+            yield self.first
+        if self.last is not None:
+            yield self.last
+
     attr_names = ()
 
 
@@ -167,7 +191,7 @@ class TypeDeclExt(c_ast.TypeDecl):
         return TypeDeclExt(td.declname, td.quals, td.type, td.coord)
 
 
-class ArrayDeclExt(c_ast.TypeDecl):
+class ArrayDeclExt(c_ast.ArrayDecl):
     @staticmethod
     def from_pycparser(ad):
         assert isinstance(ad, c_ast.ArrayDecl)
@@ -220,7 +244,6 @@ class FuncDeclExt(c_ast.Node):
 
 # {{{ attributes
 
-@template
 class _AttributesMixin(object):
     def p_attributes_opt_1(self, p):
         """ attributes_opt : attribute_decl attributes_opt
@@ -260,65 +283,8 @@ class _AttributesMixin(object):
         """
         p[0] = p[1]
 
-    # {{{ /!\ names must match C parser to override
-
-    @parameterized(('id', 'ID'), ('typeid', 'TYPEID'), ('typeid_noparen', 'TYPEID'))
-    def p_xxx_declarator_1(self, p):
-        """ xxx_declarator  : direct_xxx_declarator attributes_opt
-        """
-        if p[2].exprs:
-            if isinstance(p[1], c_ast.ArrayDecl):
-                p[1].type = to_decl_ext(p[1].type)
-                p[1].type.attributes = p[2]
-            elif isinstance(p[1], c_ast.FuncDecl):
-                p[1].type = to_decl_ext(p[1].type)
-                p[1].type.attributes = p[2]
-            elif not isinstance(p[1], c_ast.TypeDecl):
-                raise NotImplementedError(
-                        "cannot attach attributes to nodes of type '%s'"
-                        % type(p[1]))
-            else:
-                p[1] = to_decl_ext(p[1])
-                p[1].attributes = p[2]
-
-        p[0] = p[1]
-
-    # }}}
-
-    @parameterized(('id', 'ID'), ('typeid', 'TYPEID'), ('typeid_noparen', 'TYPEID'))
-    def p_xxx_declarator_2(self, p):
-        """ xxx_declarator  : pointer direct_xxx_declarator attributes_opt
-                            | pointer attributes_opt direct_xxx_declarator
-        """
-        if hasattr(p[3], "exprs"):
-            attr_decl = p[3]
-            decl = p[2]
-        else:
-            attr_decl = p[2]
-            decl = p[3]
-
-        if not attr_decl.exprs:
-            attr_decl = None
-
-        if attr_decl:
-            if isinstance(decl, c_ast.ArrayDecl):
-                decl.type = to_decl_ext(decl.type)
-                decl.type.attributes = attr_decl
-            elif isinstance(decl, c_ast.FuncDecl):
-                decl.type = to_decl_ext(decl.type)
-                decl.type.attributes = attr_decl
-            elif not isinstance(p[2], c_ast.TypeDecl):
-                raise NotImplementedError(
-                        "cannot attach attributes to nodes of type '%s'"
-                        % type(p[1]))
-            else:
-                decl = to_decl_ext(decl)
-                decl.attributes = attr_decl
-
-        p[0] = self._type_modify_decl(decl, p[1])
-
     def p_function_specifier_attr(self, p):
-        """ function_specifier  : attribute_decl
+        """ function_specifier : attribute_decl
         """
         p[0] = AttributeSpecifier(p[1])
 
@@ -364,17 +330,17 @@ class _AsmMixin(object):
         p[0] = Asm(p[1], p[3], p[5], p[7], p[9], coord=self._coord(p.lineno(2)))
 
     def p_asm_keyword(self, p):
-        """ asm_keyword : __ASM__ asm_volatile
-                        | __ASM asm_volatile
-                        | ASM asm_volatile
+        """ asm_keyword : __ASM__ asm_volatile_opt
+                        | __ASM asm_volatile_opt
+                        | ASM asm_volatile_opt
         """
         p[0] = p[1]
         if p[2]:
             p[0] += ' ' + p[2]
 
-    def p_asm_volatile(self, p):
-        """ asm_volatile : VOLATILE
-                         | empty
+    def p_asm_volatile_opt(self, p):
+        """ asm_volatile_opt : unified_volatile
+                             | empty
         """
         p[0] = p[1]
 
@@ -384,29 +350,99 @@ class _AsmMixin(object):
         """
         p[0] = p[1]
 
-    def p_statement_gnu(self, p):
-        """ statement   : asm_no_semi
-                        | asm_no_semi SEMI
+    def p_statement_asm(self, p):
+        """ statement : asm_no_semi
+                      | asm_no_semi SEMI
         """
         p[0] = p[1]
 
-#    def p_asm_with_semi(self, p):
-#        """ asm : asm SEMI
-#        """
-#        p[0] = p[1]
+    def p_asm_label_opt(self, p):
+        """ asm_label_opt : asm_keyword LPAREN unified_string_literal RPAREN
+                          | empty
+        """
+        if p[1] is None:
+            p[0] = None
+        else:
+            p[0] = Asm(p[1], p[3], None, None, None, coord=self._coord(p.lineno(2)))
+
+# }}}
 
 
 @template
 class _AsmAndAttributesMixin(_AsmMixin, _AttributesMixin):
     # {{{ /!\ names must match C parser to override
 
+    @parameterized(('id', 'ID'), ('typeid', 'TYPEID'), ('typeid_noparen', 'TYPEID'))
+    def p_xxx_declarator_1(self, p):
+        """ xxx_declarator  : direct_xxx_declarator asm_label_opt attributes_opt
+        """
+        if p[2] or p[3].exprs:
+            if isinstance(p[1], (c_ast.ArrayDecl, c_ast.FuncDecl)):
+                decl_ext = to_decl_ext(p[1].type)
+
+            elif isinstance(p[1], c_ast.TypeDecl):
+                decl_ext = to_decl_ext(p[1])
+
+            else:
+                raise NotImplementedError(
+                    "cannot attach asm or attributes to nodes of type '%s'"
+                    % type(p[1]))
+
+            if p[2]:
+                decl_ext.asm = p[2]
+
+            if p[3].exprs:
+                decl_ext.attributes = p[3]
+
+            p[1] = decl_ext
+
+        p[0] = p[1]
+
+    @parameterized(('id', 'ID'), ('typeid', 'TYPEID'), ('typeid_noparen', 'TYPEID'))
+    def p_xxx_declarator_2(self, p):
+        """ xxx_declarator  : pointer direct_xxx_declarator asm_label_opt \
+                                attributes_opt
+                            | pointer attributes_opt direct_xxx_declarator \
+                                asm_label_opt
+        """
+        if hasattr(p[4], "exprs"):
+            attr_decl = p[4]
+            asm_label = p[3]
+            decl = p[2]
+        else:
+            attr_decl = p[2]
+            asm_label = p[4]
+            decl = p[3]
+
+        if asm_label or attr_decl.exprs:
+            if isinstance(decl, (c_ast.ArrayDecl, c_ast.FuncDecl)):
+                decl_ext = to_decl_ext(decl.type)
+
+            elif isinstance(decl, c_ast.TypeDecl):
+                decl_ext = to_decl_ext(decl)
+
+            else:
+                raise NotImplementedError(
+                    "cannot attach asm or attributes to nodes of type '%s'"
+                    % type(p[1]))
+
+            if asm_label:
+                decl_ext.asm = asm_label
+
+            if attr_decl.exprs:
+                decl_ext.attributes = attr_decl
+
+            p[1] = decl_ext
+
+        p[0] = self._type_modify_decl(decl, p[1])
+
     @parameterized(('id', 'ID'), ('typeid', 'TYPEID'), ('typeid_noparen', 'TYPEID'))
     def p_direct_xxx_declarator_6(self, p):
         """ direct_xxx_declarator   : direct_xxx_declarator LPAREN parameter_type_list \
                                             RPAREN asm_opt attributes_opt
                                     | direct_xxx_declarator \
                                             LPAREN identifier_list_opt RPAREN \
-                                            asm_opt attributes_opt
+                                            asm_label_opt attributes_opt
         """
         func = FuncDeclExt(
             args=p[3],
@@ -419,7 +455,7 @@ class _AsmAndAttributesMixin(_AsmMixin, _AttributesMixin):
 
     def p_direct_abstract_declarator_6(self, p):
         """ direct_abstract_declarator  : direct_abstract_declarator \
-                LPAREN parameter_type_list_opt RPAREN asm_opt attributes_opt
+                LPAREN parameter_type_list_opt RPAREN asm_label_opt attributes_opt
         """
         func = FuncDeclExt(
             args=p[3],
@@ -452,22 +488,27 @@ class GnuCParser(_AsmAndAttributesMixin, CParserBase):
     def p_type_qualifier_gnu(self, p):
         """ type_qualifier  : __CONST
                             | __RESTRICT
+                            | __RESTRICT__
                             | __EXTENSION__
+                            | __VOLATILE
+                            | __VOLATILE__
         """
         p[0] = p[1]
 
     def p_type_specifier_gnu_typeof_expr(self, p):
         """ type_specifier  : __TYPEOF__ LPAREN expression RPAREN
+                            | TYPEOF LPAREN expression RPAREN
         """
         if isinstance(p[3], c_ast.TypeDecl):
             pass
 
-        p[0] = TypeOfExpression(p[3])
+        p[0] = TypeOfExpression(p[1], p[3])
 
     def p_type_specifier_gnu_typeof_decl(self, p):
         """ type_specifier  : __TYPEOF__ LPAREN parameter_declaration RPAREN
+                            | TYPEOF LPAREN parameter_declaration RPAREN
         """
-        p[0] = TypeOfDeclaration(p[3])
+        p[0] = TypeOfDeclaration(p[1], p[3])
 
     def p_unary_operator_gnu(self, p):
         """ unary_operator  : __REAL__
@@ -512,6 +553,26 @@ class GnuCParser(_AsmAndAttributesMixin, CParserBase):
         """ struct_declaration_list : empty """
         p[0] = None
 
+    def p_range_designator(self, p):
+        """ designator  : LBRACKET constant_expression ELLIPSIS constant_expression RBRACKET
+        """
+        p[0] = RangeExpression(p[2], p[4], coord=self._coord(p.lineno(1)))
+
+    def p_labeled_statement_4(self, p):
+        """ labeled_statement : CASE constant_expression ELLIPSIS constant_expression \
+                COLON pragmacomp_or_statement
+        """
+        p[0] = c_ast.Case(
+                RangeExpression(p[2], p[4], coord=self._coord(p.lineno(1))),
+                [p[6]],
+                self._coord(p.lineno(1)))
+
+    def p_unified_volatile_gnu(self, p):
+        """ unified_volatile : VOLATILE
+                             | __VOLATILE
+                             | __VOLATILE__
+        """
+        p[0] = p[1]
 # }}}
 
 
@@ -587,4 +648,9 @@ class OpenCLCParser(_AsmAndAttributesMixin, CParserBase):
         """
         p[0] = p[1]
 
+    def p_unified_volatile_cl(self, p):
+        """ unified_volatile : VOLATILE
+        """
+        p[0] = p[1]
+
 # vim: fdm=marker
